#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>

char buf[4096];
int alarm_time = 60;

void init_address(struct sockaddr_in *sa, const char* host, int port) {
    memset((char *) sa, 0, sizeof(*sa));
    sa->sin_family = AF_INET;
    sa->sin_port = htons(port);
    if (inet_aton(host, &sa->sin_addr)==0) {
        perror("inet_aton");
    }
}

int s1;
int s2;
struct sockaddr_in peer1;
struct sockaddr_in peer2;

int udp_socket(int port) {
    int s2;
    struct sockaddr_in si_me;

    if ((s2=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1) {
        perror("socket");
        return -1;
    }
 
    // si_me stores our local endpoint. Remember that this program
    // has to be run in a network with UDP endpoint previously known
    // and directly accessible by all clients. In simpler terms, the
    // server cannot be behind a NAT.
    memset((char *) &si_me, 0, sizeof(si_me));
    si_me.sin_family = AF_INET;
    si_me.sin_port = htons(port);
    si_me.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(s2, (struct sockaddr*)(&si_me), sizeof(si_me))==-1) {
        perror("bind");
        return -1;
    }
    return s2;
}


void alarm_handler() {
    sendto(s1, "", 0, 0, (struct sockaddr*)(&peer1), sizeof(peer1));
    //sendto(s2, "", 0, 0, (struct sockaddr*)(&peer2), sizeof(peer2));
    alarm(alarm_time);    
}

int main(int argc, char* argv[])
{
    if (argc<3) {
        fprintf(stderr, "Usage: udppair_client host1 port hostport1 hostport2 ...\n");
        return 1;
    }
    char* host1 = argv[1];
    int port1 = atoi(argv[2]);
    
    int portcount = argc-4;
    
    int ports[256];
    int sockets[256];
    struct sockaddr_in peers[256];
    int i;    
    for(i=0; i<portcount; ++i) {
        char* arg = argv[i+3];
        if(strchr(arg, ':') == NULL) {
            ports[i]=atoi(arg);
            sockets[i]=udp_socket(ports[i]);
            fcntl(sockets[i], F_SETFL, O_NONBLOCK);
            memset((char *) &peers[i], 0, sizeof(peers[i]));
        } else {
            ports[i]=0;
            sockets[i]=udp_socket(0);
            fcntl(sockets[i], F_SETFL, O_NONBLOCK);
            char* sep = strchr(arg, ':');
            *sep='\0';
            init_address(&peers[i], arg, atoi(sep+1));
        }
    }

    if(getenv("UDPPAIR_PERIOD")) {
        alarm_time = atoi(getenv("UDPPAIR_PERIOD"));
    }

    s1 = udp_socket(0);

    init_address(&peer1, host1, port1);

    fcntl(s1, F_SETFL, O_NONBLOCK);

    struct sigaction siga = {{alarm_handler}};
    sigaction(SIGALRM, &siga, NULL);
    alarm_handler();

    fd_set rfds;

    for(;;) {
select_repeat:

        FD_ZERO(&rfds);
        FD_SET(s1, &rfds);
        int maxfd = s1;
        for(i=0; i<portcount; ++i) {
            FD_SET(sockets[i], &rfds);
            if (maxfd < sockets[i]) maxfd = sockets[i];
        }
        int ret = select(maxfd+1, &rfds, NULL, NULL, NULL);

        if(ret==-1) {
            if(errno==EINTR  || errno==EAGAIN) goto select_repeat;
            perror("select");
            return 2;
        }
        
        if(FD_ISSET(s1, &rfds)) {
            socklen_t slen = sizeof(peer1);
            ret = recvfrom(s1, buf, sizeof(buf), 0, (struct sockaddr*)(&peer1), &slen);
            if(ret==-1) {
                perror("recvfrom");
                return 4;
            }
            if(ret) {
                write(1, ">", 1);
                write(1, buf, 1);
                int i = (((unsigned char)buf[0])-'0')%256;
                ret = sendto(sockets[i], buf+1, ret-1, 0, (struct sockaddr*)(&peers[i]), slen);
                if (!ret) {
                    perror("sendto");
                } else {
                }
            }
        }
        for(i=0; i<portcount; ++i) {
        if(FD_ISSET(sockets[i], &rfds)) {
            socklen_t slen = sizeof(peer2);
            struct sockaddr_in tmpaddr;
            struct sockaddr* addrptr = (struct sockaddr*)&peers[i];
            if(!ports[i]) addrptr =  (struct sockaddr*)&tmpaddr; // don't allow override this address
            ret = recvfrom(sockets[i], buf+1, sizeof(buf), 0, addrptr, &slen);
            if(ret==-1) {
                perror("recvfrom");
                return 4;
            }
            if(ret) {
                buf[0] = (i+'0')%256;
                ret = sendto(s1, buf, ret+1, 0, (struct sockaddr*)(&peer1), slen);
                if (!ret) {
                    perror("sendto");
                } else {
                    write(1, "<", 1);
                    write(1, buf, 1);
                }
            }
        }
        }
    }    
}
